#!/usr/bin/env python3

from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, FileType, Namespace
from itertools import chain
from typing import Any, Iterable

import torch
from tabulate import tabulate, tabulate_formats
from tqdm import tqdm
from transformers import (
    AutoModelForSeq2SeqLM,
    AutoTokenizer,
    M2M100ForConditionalGeneration,
)

from mbrs import timer


def buffer_lines(input_stream: Iterable[str], buffer_size: int = 64):
    buf: list[str] = []
    for i, line in enumerate(tqdm(input_stream)):
        buf.append(line.strip())
        if (i + 1) % buffer_size == 0:
            yield buf
            buf = []
    if len(buf) > 0:
        yield buf


def parse_args() -> Namespace:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    # fmt: off
    parser.add_argument("input", nargs="?", default="-",
                        type=FileType("r", encoding="utf-8"),
                        help="Input file. If not specified, read from stdin.")
    parser.add_argument("--output", "-o", default="-", type=FileType("w"),
                        help="Output file.")
    parser.add_argument("--model", "-m", type=str, default="facebook/m2m100_418M",
                        help="Model name or path.")
    parser.add_argument("--num_candidates", "-n", type=int, default=1,
                        help="Number of candidates to be returned.")
    parser.add_argument("--sampling", "-s", type=str, default="",
                        choices=["eps"],
                        help="Sampling method.")
    parser.add_argument("--beam_size", type=int, default=5,
                        help="Beam size.")
    parser.add_argument("--epsilon", "--eps", "-e", type=float, default=0.02,
                        help="Cutoff parameter for epsilon sampling.")
    parser.add_argument("--lang_pair", "-l", type=str, default="en-de",
                        help="Language name pair. Some models like M2M100 uses this information.")
    parser.add_argument("--max_length", type=int, default=1024,
                        help="Maximum length of an output sentence.")
    parser.add_argument("--min_length", type=int, default=1,
                        help="Minimum length of an output sentence.")
    parser.add_argument("--batch_size", "-b", type=int, default=8,
                        help="Batch size.")
    parser.add_argument("--sampling_size", type=int, default=8,
                        help="Sampling size in a single inference.")
    parser.add_argument("--fp16", action="store_true",
                        help="Use float16.")
    parser.add_argument("--bf16", action="store_true",
                        help="Use bfloat16.")
    parser.add_argument("--cpu", action="store_true",
                        help="Force to use CPU.")
    parser.add_argument("--quiet", "-q", action="store_true",
                        help="No report statistics.")
    parser.add_argument("--report_format", type=str, default="rounded_outline",
                        choices=tabulate_formats,
                        help="Report runtime statistics.")
    parser.add_argument("--width", "-w", type=int, default=1,
                        help="Number of digits for values of float point.")
    # fmt: on
    return parser.parse_args()


def main(args: Namespace) -> None:
    src_lang, tgt_lang = tuple(args.lang_pair.split("-"))
    tokenizer = AutoTokenizer.from_pretrained(args.model)
    model = AutoModelForSeq2SeqLM.from_pretrained(args.model)

    model.eval()
    for param in model.parameters():
        param.requires_grad = False
    if torch.cuda.is_available() and not args.cpu:
        model.cuda()
        if args.fp16:
            model.half()
        elif args.bf16:
            model.bfloat16()

    generation_kwargs = {"max_length": args.max_length, "min_length": args.min_length}

    if isinstance(model, M2M100ForConditionalGeneration):
        tokenizer.src_lang = src_lang
        generation_kwargs["forced_bos_token_id"] = tokenizer.get_lang_id(tgt_lang)

    if args.sampling == "eps":
        generation_kwargs["do_sample"] = True
        generation_kwargs["epsilon_cutoff"] = args.epsilon
        generation_kwargs["num_beams"] = 1
    else:
        generation_kwargs["num_beams"] = max(args.beam_size, args.num_candidates)

    def decode(
        inputs: list[str], num_candidates: int, generation_kwargs: dict[str, Any]
    ) -> list[str]:
        model_inputs = tokenizer(inputs, return_tensors="pt", padding=True).to(
            device=model.device
        )
        with timer.measure("generate"):
            model_outputs = model.generate(
                **model_inputs, **generation_kwargs, num_return_sequences=num_candidates
            )
        return tokenizer.batch_decode(model_outputs, skip_special_tokens=True)

    def generate(inputs: list[str]) -> list[str]:
        if (
            not generation_kwargs.get("do_sample", False)
            or generation_kwargs.get("num_beams", 1) != 1
        ):
            return decode(inputs, args.num_candidates, generation_kwargs)
        else:
            outputs: list[list[str]] = [[] for _ in range(args.batch_size)]
            for n in range(0, args.num_candidates, args.sampling_size):
                sampling_size = min(args.sampling_size, args.num_candidates - n)
                samples = decode(inputs, sampling_size, generation_kwargs)
                for i in range(args.batch_size):
                    outputs[i] += samples[i * sampling_size : (i + 1) * sampling_size]
            return list(chain.from_iterable(outputs))

    num_sentences = 0
    for lines in buffer_lines(args.input, buffer_size=args.batch_size):
        for output in generate(lines):
            print(output.strip(), file=args.output)
            num_sentences += 1

    if not args.quiet:
        statistics = timer.aggregate().result(num_sentences)
        table = tabulate(
            statistics,
            headers="keys",
            tablefmt=args.report_format,
            floatfmt=f".{args.width}f",
        )
        print(table)


def cli_main():
    args = parse_args()
    main(args)


if __name__ == "__main__":
    cli_main()



import pytest
import torch

from mbrs.metrics import get_metric

from .mbr import DecoderMBR

HYPOTHESES = [
    ["another test", "this is a test", "this is an test", "this is a fest"],
    ["another test", "this is a fest", "this is a test"],
    ["this is a test"],
    ["Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;"],
]
REFERENCES = [
    ["another test", "this is a test", "this is an test", "this is a fest"],
    ["this is a test", "ref1 ref2 ref3 ref4"],
    ["this is a test"],
    ["producţia de zahăr brut se exprimă în zahăr alb;"],
]

BEST_INDICES = [1, 2, 0, 0]
BEST_SENTENCES = [
    "this is a test",
    "this is a test",
    "this is a test",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]

SCORES = {
    "bleu": torch.Tensor([52.697, 50.0, 100.0, 8.493]),
    "ter": torch.Tensor([50.000, 50.000, 0.000, 100.000]),
}


class TestDecoderMBR:
    @pytest.mark.parametrize("metric_type", ["bleu", "ter"])
    def test_decode(self, metric_type: str):
        metric_cls = get_metric(metric_type)
        decoder = DecoderMBR(DecoderMBR.Config(), metric_cls(metric_cls.Config()))
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, nbest=1)
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]
            torch.testing.assert_close(
                torch.tensor(output.score[0]),
                SCORES[metric_type][i],
                atol=0.0005,
                rtol=1e-4,
            )



import torch

from .chrf import MetricChrF

HYPOTHESES = [
    "this is a test",
    "another test",
    "this is a fest",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
REFERENCES = [
    "ref",
    "this is a test",
    "producţia de zahăr brut se exprimă în zahăr alb;",
]
SCORES = torch.Tensor(
    [
        [7.246, 100.0, 3.411],
        [14.493, 23.060, 4.386],
        [14.493, 60.628, 3.411],
        [6.849, 7.224, 46.161],
    ],
)


class TestMetricChrF:
    def test_score(self):
        metric = MetricChrF(MetricChrF.Config())
        for i, hyp in enumerate(HYPOTHESES):
            for j, ref in enumerate(REFERENCES):
                assert torch.isclose(
                    SCORES[i, j], torch.tensor(metric.score(hyp, ref)), atol=0.0005
                )

    def test_expected_scores(self):
        metric = MetricChrF(MetricChrF.Config())
        expected_scores = metric.expected_scores(HYPOTHESES, REFERENCES)
        torch.testing.assert_close(
            expected_scores,
            SCORES.mean(dim=1),
            atol=0.0005,
            rtol=1e-4,
        )



import pytest
import torch

from mbrs.metrics.comet import MetricCOMET

from .cbmbr import DecoderCBMBR

SOURCE = [
    "これはテストです",
    "これはテストです",
    "これはテストです",
    "これはテストです",
]
HYPOTHESES = [
    ["another test", "this is a test", "this is a fest"],
    ["another test", "this is a fest", "this is a test"],
    ["this is a test"],
    ["Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;"],
]
REFERENCES = [
    ["another test", "this is a test", "this is a fest"],
    ["this is a test", "ref", "these are tests", "this is the test"],
    ["this is a test"],
    ["producţia de zahăr brut se exprimă în zahăr alb;"],
]

BEST_INDICES = [1, 2, 0, 0]
BEST_SENTENCES = [
    "this is a test",
    "this is a test",
    "this is a test",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
SCORES = torch.Tensor([0.88974, 0.76127, 0.99257, 0.78060])

NCENTROIDS = 2


class TestDecoderCBMBR:
    @pytest.mark.parametrize("kmeanspp", [True, False])
    def test_decode(self, metric_comet: MetricCOMET, kmeanspp: bool):
        decoder = DecoderCBMBR(
            DecoderCBMBR.Config(ncentroids=NCENTROIDS, kmeanspp=kmeanspp), metric_comet
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, SOURCE[i], nbest=1)
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]
            assert torch.isclose(
                torch.tensor(output.score[0]),
                SCORES[i],
                atol=0.0005 / 100,
            )



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Optional

from mbrs.metrics.base import Metric, MetricBase, MetricReferenceless


class DecoderBase(abc.ABC):
    """Decoder base class."""

    def __init__(self, cfg: DecoderBase.Config, metric: MetricBase) -> None:
        self.cfg = cfg
        self.metric = metric

    @dataclass
    class Config:
        """Configuration for the decoder."""

    @dataclass
    class Output:
        """
        - idx (list[int]): Index numbers of the n-best hypotheses.
        - sentence (list[str]): Sentences of the n-best hypotheses.
        - score (list[float]): Scores of the n-best hypotheses.
        """

        idx: list[int]
        sentence: list[str]
        score: list[float]


class DecoderReferenceBased(DecoderBase, metaclass=abc.ABCMeta):
    """Decoder base class for strategies that use references like MBR decoding."""

    metric: Metric

    @abc.abstractmethod
    def decode(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        nbest: int = 1,
    ) -> DecoderReferenceBased.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            nbest (int): Return the n-best hypotheses.

        Returns:
            Decoder.Output: The n-best hypotheses.
        """


class DecoderReferenceless(DecoderBase, metaclass=abc.ABCMeta):
    """Decoder base class for reference-free strategies."""

    metric: MetricReferenceless

    @abc.abstractmethod
    def decode(
        self, hypotheses: list[str], source: str, nbest: int = 1
    ) -> DecoderReferenceless.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            source (str): A source.
            nbest (int): Return the n-best hypotheses.

        Returns:
            Decoder.Output: The n-best hypotheses.
        """



import torch

from mbrs.metrics.cometqe import MetricCOMETQE

from .rerank import DecoderRerank

SOURCE = "これはテストです"
HYPOTHESES = [
    ["another test", "this is a test", "this is a fest"],
    ["another test", "this is a fest", "this is a test"],
    ["this is a test"],
    ["Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;"],
]

BEST_INDICES = [1, 2, 0, 0]
BEST_SENTENCES = [
    "this is a test",
    "this is a test",
    "this is a test",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
SCORES = torch.Tensor([0.86415, 0.86415, 0.86415, 0.29771])


class TestDecoderRerank:
    def test_decode(self, metric_cometqe: MetricCOMETQE):
        decoder = DecoderRerank(DecoderRerank.Config(), metric_cometqe)
        for i, hyps in enumerate(HYPOTHESES):
            output = decoder.decode(hyps, SOURCE, nbest=1)
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]
            torch.testing.assert_close(
                torch.tensor(output.score[0]),
                SCORES[i],
                atol=0.0005 / 100,
                rtol=1e-6,
            )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from sacrebleu.metrics.bleu import BLEU

from . import Metric, register


@register("bleu")
class MetricBLEU(Metric):
    """BLEU metric class."""

    @dataclass
    class Config(Metric.Config):
        """BLEU metric configuration.

        - lowercase (bool): If True, lowercased BLEU is computed.
        - force (bool): Ignore data that looks already tokenized.
        - tokenize (str, optional): The tokenizer to use. If None, defaults to language-specific tokenizers with '13a' as the fallback default.
        - smooth_method (str): The smoothing method to use ('floor', 'add-k', 'exp' or 'none').
        - smooth_value (float, optional): The smoothing value for `floor` and `add-k` methods. `None` falls back to default value.
        - max_ngram_order (int): If given, it overrides the maximum n-gram order (default: 4) when computing precisions.
        - effective_order (bool): If `True`, stop including n-gram orders for which precision is 0.
            This should be `True`, if sentence-level BLEU will be computed. (default: True)
        - trg_lang (str): An optional language code to raise potential tokenizer warnings.
        """

        lowercase: bool = False
        force: bool = False
        tokenize: Optional[str] = None
        smooth_method: str = "exp"
        smooth_value: Optional[float] = None
        max_ngram_order: int = 4
        effective_order: bool = True
        trg_lang: str = ""

    def __init__(self, cfg: MetricBLEU.Config):
        self.scorer = BLEU(
            lowercase=cfg.lowercase,
            force=cfg.force,
            tokenize=cfg.tokenize,
            smooth_method=cfg.smooth_method,
            smooth_value=cfg.smooth_value,
            max_ngram_order=cfg.max_ngram_order,
            effective_order=cfg.effective_order,
            trg_lang=cfg.trg_lang,
        )

    def score(self, hypothesis: str, reference: str, *_) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): Hypothesis.
            reference (str): Reference.

        Returns:
            float: The score of the given hypothesis.
        """
        return self.scorer.sentence_score(hypothesis, [reference]).score



from mbrs import registry

from .base import Metric, MetricCacheable, MetricReferenceless

register, get_metric = registry.setup("metric")

from .bleu import MetricBLEU
from .chrf import MetricChrF
from .comet import MetricCOMET
from .cometqe import MetricCOMETQE
from .ter import MetricTER

__all__ = [
    "Metric",
    "MetricCacheable",
    "MetricReferenceless",
    "MetricBLEU",
    "MetricChrF",
    "MetricCOMET",
    "MetricCOMETQE",
    "MetricTER",
]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

import torch

from mbrs import timer
from mbrs.metrics import MetricCacheable

from . import register
from .mbr import DecoderMBR


@register("pruning_mbr")
class DecoderPruningMBR(DecoderMBR):
    """Pruning MBR decoder class.

    References:
        J. Cheng and A. Vlachos, 2023,
        "Faster Minimum Bayes Risk Decoding with Confidence-based Pruning".
        https://aclanthology.org/2023.emnlp-main.767/
    """

    cfg: Config

    @dataclass
    class Config(DecoderMBR.Config):
        """Configuration for the decoder.

        - alpha (float): Prune hypotheses based on this confidence threshold.
        - sampling_shceduler (list[int]): Sample size scheduler. For each step, the
          number of samples will be the t-th number.
        - num_boostrap_samples (int): Number of boostrap samples.
        - seed (int): Random seed for bootstrap sampling. The random numbers are
          generated using PCG-64.
        """

        alpha: float = 0.99
        sampling_scheduler: list[int] = field(
            default_factory=lambda: [8, 16, 32, 64, 128, 256]
        )
        num_bootstrap_samples: int = 500
        seed: int = 0

    def decode_pruning(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        nbest: int = 1,
    ) -> tuple[list[float], list[int]]:
        """Select the n-best hypotheses using pruning MBR decoding.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            nbest (int): Return the n-best hypotheses.

        Returns:
            - list[float]: Top-k scores.
            - list[int]: Top-k indices.
        """
        rng = torch.Generator(device=self.metric.device).manual_seed(self.cfg.seed)
        H = len(hypotheses)
        max_r = min(max(self.cfg.sampling_scheduler), len(references))
        pairwise_scores = torch.zeros((H, max_r), device=self.metric.device)
        orig_indices = torch.arange(H, device=self.metric.device)

        if isinstance(self.metric, MetricCacheable):
            with timer.measure("encode/hypotheses"):
                hypotheses_ir = self.metric.encode(hypotheses)
            references_ir = hypotheses_ir if hypotheses == references else None
            if source is None:
                source_ir = None
            else:
                with timer.measure("encode/source"):
                    source_ir = self.metric.encode([source])

        with timer.measure("pruning_mbr"):
            # Algorithm 1 in the paper.
            prev_r = 0
            for t, r in enumerate(self.cfg.sampling_scheduler):
                r = min(r, len(references))
                if r <= prev_r:
                    break

                # Equation 5 and Algorithm 2 in the paper.
                if isinstance(self.metric, MetricCacheable):
                    if references_ir is None:
                        with timer.measure("encode/references"):
                            references_ir_t = self.metric.encode(references[prev_r:r])
                    else:
                        references_ir_t = references_ir[prev_r:r]

                    pairwise_scores[:, prev_r:r] = self.metric.pairwise_scores_from_ir(
                        hypotheses_ir, references_ir_t, source_ir
                    )
                else:
                    pairwise_scores[:, prev_r:r] = self.metric.pairwise_scores(
                        hypotheses, references[prev_r:r], source
                    )
                expected_scores = pairwise_scores.mean(dim=-1)
                current_best_idx = self.metric.argbest(expected_scores)
                sample_indices = torch.randint(
                    r,
                    size=(self.cfg.num_bootstrap_samples, r),
                    device=self.metric.device,
                    generator=rng,
                )
                bootstrap_expected_scores = pairwise_scores[:, sample_indices].mean(
                    dim=-1
                )

                win_rates = (
                    (
                        bootstrap_expected_scores
                        >= bootstrap_expected_scores[current_best_idx]
                    )
                    .float()
                    .mean(dim=1)
                )
                winners = (win_rates > 1 - self.cfg.alpha).nonzero(as_tuple=True)[0]
                num_winners = len(winners)
                if num_winners >= nbest:
                    if isinstance(self.metric, MetricCacheable):
                        hypotheses_ir = hypotheses_ir.index_select(0, winners)
                    else:
                        hypotheses = [hypotheses[i] for i in winners]
                    pairwise_scores = pairwise_scores[winners]
                    orig_indices = orig_indices[winners]
                    prev_r = r
                else:
                    break
            expected_scores = pairwise_scores[:, :prev_r].mean(dim=1)

        topk_scores, topk_indices = self.metric.topk(expected_scores, k=nbest)
        return topk_scores, orig_indices[topk_indices].tolist()

    def decode(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        nbest: int = 1,
    ) -> DecoderMBR.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            nbest (int): Return the n-best hypotheses.

        Returns:
            DecoderMBR.Output: The n-best hypotheses.
        """

        if len(hypotheses) <= nbest:
            expected_scores = self.metric.expected_scores(
                hypotheses, references, source
            )
            topk_scores, topk_indices = self.metric.topk(expected_scores, k=nbest)
        else:  # Pruning MBR decoding
            topk_scores, topk_indices = self.decode_pruning(
                hypotheses, references, source
            )

        return self.Output(
            idx=topk_indices,
            sentence=[hypotheses[idx] for idx in topk_indices],
            score=topk_scores,
        )



"""
Timer module including some useful global objects.

Example:
    >>> from mbrs import timer
    >>> for hyps, src in zip(hypotheses, sources):
            with timer.measure("encode/hypotheses") as t:
                h = metric.encode(hyps)
            t.set_delta_ncalls(len(hyps))
            with timer.measure("encode/source"):
                s = metric.encode([src])
            with timer.measure("score"):
                scores = metric.score(h, s)
    >>> res = timer.aggregate().result()  # return the result table
"""

from __future__ import annotations

import contextlib
import time
from collections import defaultdict
from dataclasses import dataclass, field


class Stopwatch:
    """Stopwatch class to measure the elapsed time.

    Example:
        >>> timer = Stopwatch()
        >>> for i in range(10):
                with timer():
                    time.sleep(1)
        >>> print(f"{timer.elapsed_time:.3f}")
        10.000

        >>> timer = Stopwatch()
        >>> for i in range(10):
                with timer() as t:
                    time.sleep(2)
                    t.set_delta_ncalls(2)
        >>> print(f"{timer.elapsed_time:.3f}")
        20.000
        >>> print(f"{timer.ncalls}")
        20
    """

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        """Reset the stopwatch."""
        self._acc_time = 0.0
        self._acc_ncalls = 0
        self._delta_ncalls = 1

    @contextlib.contextmanager
    def __call__(self):
        """Measure the time."""
        _acc_time = self._acc_time
        _acc_ncalls = self._acc_ncalls
        start = time.perf_counter()
        try:
            self.set_delta_ncalls(1)  # Set to 1 in the default
            yield self
        finally:
            # Treat nest calling
            if _acc_time == self._acc_time:
                self._acc_time = _acc_time + time.perf_counter() - start
            if _acc_ncalls == self._acc_ncalls:
                self._acc_ncalls = _acc_ncalls + self._delta_ncalls

    def set_delta_ncalls(self, delta: int = 1):
        """Set delta for counting the number of calls."""
        self._delta_ncalls = delta

    @property
    def elpased_time(self) -> float:
        """Return the total elapsed time."""
        return self._acc_time

    @property
    def ncalls(self) -> int:
        """Return the number of calls."""
        return self._acc_ncalls


class StopwatchDict(defaultdict[str, Stopwatch]):
    """A dictionary of the :class:`Stopwatch` class.

    Example:
        >>> timers = StopwatchDict()
        >>> for i in range(10):
                with timers("A"):
                    time.sleep(1)
            for i in range(3):
                with timers("B"):
                    time.sleep(1)
        >>> print(f"{timers.total}")
        {"A": 10.000, "B": 3.000}
    """

    def __init__(self) -> None:
        super().__init__(Stopwatch)

    def reset(self) -> None:
        """Reset all stopwatches."""
        for t in self.values():
            t.reset()

    @contextlib.contextmanager
    def __call__(self, name: str):
        """Measure the time."""
        with self[name]() as timer:
            try:
                yield timer
            finally:
                pass

    @property
    def elapsed_time(self) -> dict[str, float]:
        """Return the total elapsed time."""
        return {k: v.elpased_time for k, v in self.items()}

    @property
    def ncalls(self) -> dict[str, int]:
        """Return the number of calls."""
        return {k: v.ncalls for k, v in self.items()}


measure = StopwatchDict()


@dataclass
class ProfileTree:
    elapsed_time: float = -1.0
    ncalls: int = -1
    children: dict[str, ProfileTree] = field(default_factory=dict)

    @property
    def is_leaf(self) -> bool:
        """Return whether the node is leaf or not."""
        return len(self.children) == 0

    def aggregate(self):
        if self.is_leaf:
            if self.elapsed_time < 0.0:
                raise RuntimeError("Missing elapsed_time")
            if self.ncalls < 0:
                raise RuntimeError("Missing ncalls")
        else:
            for child in self.children.values():
                child.aggregate()
            if self.elapsed_time < 0.0:
                self.elapsed_time = 0.0
                for child in self.children.values():
                    self.elapsed_time += child.elapsed_time
            if self.ncalls < 0:
                self.ncalls = 0
                for child in self.children.values():
                    self.ncalls += child.ncalls

    @classmethod
    def build(cls, timers: StopwatchDict, separetor: str = "/"):
        root = cls()
        for name, timer in timers.items():
            prefix = name.split(separetor)
            node = root
            for path in prefix:
                if path not in node.children:
                    node.children[path] = ProfileTree()
                node = node.children[path]
            node.elapsed_time = timer.elpased_time
            node.ncalls = timer.ncalls
        root.aggregate()
        return root

    def result(self, nsentences: int = -1) -> list[dict[str, str | int | float]]:
        def _result(name: str, node: ProfileTree) -> list[dict[str, str | int | float]]:
            stat = {
                "name": name.strip("/"),
                "acctime": node.elapsed_time,
                "acccalls": node.ncalls,
                "ms/call": node.elapsed_time * 1000 / node.ncalls,
            }
            if nsentences > 0:
                stat["ms/sentence"] = node.elapsed_time * 1000 / nsentences
                stat["calls/sentence"] = node.ncalls / nsentences
            res = [stat]
            for path, child in node.children.items():
                res += _result(name + "/" + path, child)
            return res

        res = _result("", self)
        return res[1:]  # Remove the root node


def aggregate() -> ProfileTree:
    """Aggregate the timers.

    Returns:
        ProfileTree: The root of the profile tree.
    """
    return ProfileTree.build(measure)



from __future__ import annotations

from mbrs import timer

from . import DecoderReferenceless, register


@register("rerank")
class DecoderRerank(DecoderReferenceless):
    """Reranking decoder class.

    - Time complexity: O(N)
    - Space complexity: O(N)
    """

    def decode(
        self, hypotheses: list[str], source: str, nbest: int = 1
    ) -> DecoderRerank.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            source (str): A source.
            nbest (int): Return the n-best hypotheses.

        Returns:
            DecoderRerank.Output: The n-best hypotheses.
        """
        with timer.measure("rerank"):
            scores = self.metric.scores(hypotheses, source)
        topk_scores, topk_indices = self.metric.topk(scores, k=nbest)
        return self.Output(
            idx=topk_indices,
            sentence=[hypotheses[idx] for idx in topk_indices],
            score=topk_scores,
        )



from __future__ import annotations

from dataclasses import dataclass

import torch
from comet import download_model, load_from_checkpoint
from torch import Tensor

from . import MetricCacheable, register


@register("comet")
class MetricCOMET(MetricCacheable):
    """COMET metric class."""

    @dataclass
    class Config(MetricCacheable.Config):
        """COMET metric configuration.

        - model (str): Model name or path.
        - batch_size (int): Batch size.
        - fp16 (bool): Use float16 for the forward computation.
        - bf16 (bool): Use bfloat16 for the forward computation.
        - cpu (bool): Use CPU for the forward computation.
        """

        model: str = "Unbabel/wmt22-comet-da"
        batch_size: int = 64
        fp16: bool = False
        bf16: bool = False
        cpu: bool = False

    def __init__(self, cfg: MetricCOMET.Config):
        self.cfg = cfg
        self.scorer = load_from_checkpoint(download_model(cfg.model))
        self.scorer.eval()
        for param in self.scorer.parameters():
            param.requires_grad = False

        if not cfg.cpu and torch.cuda.is_available():
            self.scorer = self.scorer.cuda()
            if cfg.fp16:
                self.scorer = self.scorer.half()
            elif cfg.bf16:
                self.scorer = self.scorer.bfloat16()

    @property
    def device(self) -> torch.device:
        """Returns the device of the model."""
        return self.scorer.device

    def encode(self, sentences: list[str]) -> torch.Tensor:
        """Compute sentence embedding vectors of the given sentences.

        Args:
            sentences (list[str]): Input sentences.

        Returns:
            torch.Tensor: Sentence embeddings of shape `(N, D)`, where
              - N: the number of sentences
              - D: size of the embedding dimmension
        """
        batches = [
            self.scorer.encoder.prepare_sample(sentences[i : i + self.cfg.batch_size])
            for i in range(0, len(sentences), self.cfg.batch_size)
        ]
        embeds = []
        for batch in batches:
            emb = self.scorer.get_sentence_embedding(**batch.to(self.scorer.device))
            if self.scorer.device.type != "cpu":
                if self.cfg.fp16:
                    emb = emb.half()
                elif self.cfg.bf16:
                    emb = emb.bfloat16()
                else:
                    emb = emb.float()
            embeds.append(emb)
        embeds = torch.vstack(embeds)
        return embeds

    def out_proj(
        self, hypotheses_ir: Tensor, references_ir: Tensor, source_ir: Tensor
    ) -> Tensor:
        """Forward the output projection layer.

        Args:
            hypotheses_ir (Tensor): Intermediate representations of hypotheses.
            references_ir (Tensor): Intermediate representations of references.
            source_ir (Tensor): Intermediate representations of a source.

        Returns:
            Tensor: N scores.
        """
        return self.scorer.estimate(source_ir, hypotheses_ir, references_ir)["score"]



import numpy as np

from mbrs.metrics.chrf import MetricChrF
from mbrs.metrics.comet import MetricCOMET

from .pruning_mbr import DecoderPruningMBR

SOURCE = [
    "これはテストです",
    "これはテストです",
    "これはテストです",
    "これはテストです",
]
HYPOTHESES = [
    ["another test", "this is a test", "this is a fest", "x", "this is test"],
    ["another test", "this is a fest", "this is a test"],
    ["this is a test"],
    ["Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;"],
]
REFERENCES = [
    ["another test", "this is a test", "this is a fest", "x", "this is test"],
    ["this is a test", "ref", "these are tests", "this is the test"],
    ["this is a test"],
    ["producţia de zahăr brut se exprimă în zahăr alb;"],
]

BEST_INDICES = [1, 2, 0, 0]
BEST_SENTENCES = [
    "this is a test",
    "this is a test",
    "this is a test",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
SCORES_COMET = np.array([0.84780, 0.85304, 0.99257, 0.78060], dtype=np.float32)
SCORES_CHRF = np.array([48.912, 44.239, 100.0, 46.161], dtype=np.float32)


class TestDecoderPruningMBR:
    def test_decode_chrf(self):
        metric = MetricChrF(MetricChrF.Config())
        decoder = DecoderPruningMBR(
            DecoderPruningMBR.Config(sampling_scheduler=[2, 4, 8]), metric
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, SOURCE[i], nbest=1)
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]
            assert np.allclose(
                np.array(output.score[0], dtype=np.float32), SCORES_CHRF[i], atol=0.0005
            )

    def test_decode_comet(self, metric_comet: MetricCOMET):
        decoder = DecoderPruningMBR(
            DecoderPruningMBR.Config(sampling_scheduler=[2, 4, 8]), metric_comet
        )
        for i, (hyps, refs) in enumerate(zip(HYPOTHESES, REFERENCES)):
            output = decoder.decode(hyps, refs, SOURCE[i], nbest=1)
            assert output.idx[0] == BEST_INDICES[i]
            assert output.sentence[0] == BEST_SENTENCES[i]
            assert np.allclose(
                np.array(output.score[0], dtype=np.float32),
                SCORES_COMET[i],
                atol=0.0005,
            )



from __future__ import annotations

from dataclasses import dataclass

import torch
from comet import download_model, load_from_checkpoint
from transformers import BatchEncoding

from . import MetricReferenceless, register


@register("cometqe")
class MetricCOMETQE(MetricReferenceless):
    """COMET-QE metric class."""

    @dataclass
    class Config(MetricReferenceless.Config):
        """COMET-QE metric configuration.

        - model (str): Model name or path.
        - batch_size (int): Batch size.
        - fp16 (bool): Use float16 for the forward computation.
        - bf16 (bool): Use bfloat16 for the forward computation.
        - cpu (bool): Use CPU for the forward computation.
        """

        model: str = "Unbabel/wmt22-cometkiwi-da"
        batch_size: int = 64
        fp16: bool = False
        bf16: bool = False
        cpu: bool = False

    def __init__(self, cfg: MetricCOMETQE.Config):
        self.cfg = cfg
        self.scorer = load_from_checkpoint(download_model(cfg.model))
        self.scorer.eval()
        for param in self.scorer.parameters():
            param.requires_grad = False

        if not cfg.cpu and torch.cuda.is_available():
            self.scorer = self.scorer.cuda()
            if cfg.fp16:
                self.scorer = self.scorer.half()
            elif cfg.bf16:
                self.scorer = self.scorer.bfloat16()

    @property
    def device(self) -> torch.device:
        """Returns the device of the model."""
        return self.scorer.device

    def score(self, hypothesis: str, source: str) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): A hypothesis.
            source (str): A source.

        Returns:
            float: The score of the given hypothesis.
        """
        return self.scores([hypothesis], source).item()

    def scores(self, hypotheses: list[str], source: str) -> torch.Tensor:
        """Calculate the scores of hypotheses.

        Args:
            hypotheses (list[str]): Hypotheses.
            source (str): A source.

        Returns:
            torch.Tensor: The scores of hypotheses.
        """
        data = [{"src": source, "mt": hyp} for hyp in hypotheses]
        scores = []
        for i in range(0, len(data), self.cfg.batch_size):
            batch = BatchEncoding(
                self.scorer.prepare_for_inference(data[i : i + self.cfg.batch_size])[0]
            ).to(self.scorer.device)
            model_output = self.scorer.predict_step((batch,))
            scores.append(model_output.scores)
        return torch.cat(scores)



import pytest

from mbrs.metrics import MetricCOMET, MetricCOMETQE


@pytest.fixture(scope="session")
def metric_comet():
    return MetricCOMET(MetricCOMET.Config())


@pytest.fixture(scope="session")
def metric_cometqe():
    return MetricCOMETQE(MetricCOMETQE.Config())



from mbrs import registry

from .base import DecoderReferenceBased, DecoderReferenceless

register, get_decoder = registry.setup("decoder")

from .cbmbr import DecoderCBMBR
from .mbr import DecoderMBR
from .pruning_mbr import DecoderPruningMBR
from .rerank import DecoderRerank

__all__ = [
    "DecoderReferenceBased",
    "DecoderReferenceless",
    "DecoderMBR",
    "DecoderCBMBR",
    "DecoderPruningMBR",
    "DecoderRerank",
]



from __future__ import annotations

from dataclasses import dataclass

from sacrebleu.metrics.chrf import CHRF

from . import Metric, register


@register("chrf")
class MetricChrF(Metric):
    """ChrF metric class."""

    @dataclass
    class Config(Metric.Config):
        """ChrF metric configuration.

        - char_order (int): Character n-gram order.
        - word_order (int): Word n-gram order. If equals to 2, the metric is referred to as chrF++.
        - beta (int): Determine the importance of recall w.r.t precision.
        - lowercase (bool): Enable case-insensitivity.
        - whitespace (bool): If `True`, include whitespaces when extracting character n-grams.
        - eps_smoothing (bool): If `True`, applies epsilon smoothing similar to reference chrF++.py, NLTK and Moses implementations.
            Otherwise, it takes into account effective match order similar to sacreBLEU < 2.0.0.
        """

        char_order: int = 6
        word_order: int = 0
        beta: int = 2
        lowercase: bool = False
        whitespace: bool = False
        eps_smoothing: bool = False

    def __init__(self, cfg: MetricChrF.Config):
        self.scorer = CHRF(
            char_order=cfg.char_order,
            word_order=cfg.word_order,
            beta=cfg.beta,
            lowercase=cfg.lowercase,
            whitespace=cfg.whitespace,
            eps_smoothing=cfg.eps_smoothing,
        )

    def score(self, hypothesis: str, reference: str, *_) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): Hypothesis.
            reference (str): Reference.

        Returns:
            float: The score of the given hypothesis.
        """
        return self.scorer.sentence_score(hypothesis, [reference]).score



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Optional

import torch
from torch import Tensor

from mbrs import timer


class MetricBase(abc.ABC):
    """Base metric class."""

    def __init__(self, cfg: Config):
        self.cfg = cfg

    HIGHER_IS_BETTER: bool = True

    @dataclass
    class Config: ...

    @property
    def device(self) -> torch.device:
        """Returns the device of the metric object."""
        return torch.device("cpu")

    def topk(self, x: Tensor, k: int = 1) -> tuple[list[float], list[int]]:
        """Return the top-k best elements and corresponding indices.

        Args:
            x (Tensor): Input 1-D array.
            k (int): Return the top-k values and indices.

        Returns:
            tuple[list[float], list[int]]
              - list[float]: The top-k values.
              - list[int]: The top-k indices.
        """
        values, indices = torch.topk(x, k=min(k, len(x)), largest=self.HIGHER_IS_BETTER)
        return values.tolist(), indices.tolist()

    def argbest(self, x: Tensor) -> Tensor:
        """Return the index of the best element.

        Args:
            x (Tensor): Input 1-D array.

        Returns:
            Tensor: A scalar tensor of the best index.
        """
        if self.HIGHER_IS_BETTER:
            return torch.argmax(x)
        return torch.argmin(x)


class Metric(MetricBase, metaclass=abc.ABCMeta):
    """Base metric class."""

    @abc.abstractmethod
    def score(
        self, hypothesis: str, reference: str, source: Optional[str] = None
    ) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): A hypothesis.
            reference (str): A reference.
            source (str, optional): A source.

        Returns:
            float: The score of the given hypothesis.
        """

    def pairwise_scores(
        self, hypotheses: list[str], references: list[str], source: Optional[str] = None
    ) -> Tensor:
        """Calculate the pairwise scores.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.

        Returns:
            Tensor: Score matrix of shape `(H, R)`, where `H` is the number
              of hypotheses and `R` is the number of references.
        """
        with timer.measure("score") as t:
            t.set_delta_ncalls(len(hypotheses) * len(references))
            return Tensor(
                [
                    [self.score(hyp, ref, source) for ref in references]
                    for hyp in hypotheses
                ]
            )

    def expected_scores(
        self, hypotheses: list[str], references: list[str], source: Optional[str] = None
    ) -> Tensor:
        """Calculate the expected scores for each hypothesis.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.

        Returns:
            Tensor: The expected scores for each hypothesis.
        """
        with timer.measure("expectation"):
            return self.pairwise_scores(hypotheses, references, source).mean(dim=1)


class MetricCacheable(Metric, metaclass=abc.ABCMeta):
    """Base class for cacheable metrics.

    This class supports to cache intermediate representations of the encoder."""

    @abc.abstractmethod
    def encode(self, sentences: list[str]) -> Tensor:
        """Encode the given sentences into their intermediate representations.

        Args:
            sentences (list[str]): Input sentences.

        Returns:
            Tensor: Intermediate representations of shape `(N, D)` where `N` is the
              number of hypotheses and `D` is a size of the embedding dimension.
        """

    @abc.abstractmethod
    def out_proj(
        self,
        hypotheses_ir: Tensor,
        references_ir: Tensor,
        source_ir: Optional[Tensor] = None,
    ) -> Tensor:
        """Forward the output projection layer.

        Args:
            hypotheses_ir (Tensor): Intermediate representations of hypotheses.
            references_ir (Tensor): Intermediate representations of references.
            source_ir (Tensor, optional): Intermediate representations of a source.

        Returns:
            Tensor: N scores.
        """

    def score(
        self,
        hypothesis: str,
        reference: str,
        source: Optional[str] = None,
    ) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): A hypothesis.
            reference (str): A reference.
            source (str, optional): A source.

        Returns:
            float: The score of the given hypothesis.
        """
        return self.out_proj(
            self.encode([hypothesis]),
            self.encode([reference]),
            self.encode([source]) if source is not None else None,
        ).item()

    def pairwise_scores_from_ir(
        self,
        hypotheses_ir: Tensor,
        references_ir: Tensor,
        source_ir: Optional[Tensor] = None,
    ) -> Tensor:
        """Calculate the pairwise scores from the intermediate representations.

        Args:
            hypotheses_ir (Tensor): Hypotheses.
            references_ir (Tensor): References.
            source_ir (Tensor, optional): A source.

        Returns:
            Tensor: Score matrix of shape `(H, R)`, where `H` is the number
              of hypotheses and `R` is the number of references.
        """
        H, D = hypotheses_ir.size()
        R, _ = references_ir.size()
        if source_ir is not None:
            source_ir = source_ir.repeat(H, 1)

        scores = []
        for i in range(R):
            with timer.measure("score") as t:
                t.set_delta_ncalls(H)
                scores.append(
                    self.out_proj(
                        hypotheses_ir, references_ir[i, :].repeat(H, 1), source_ir
                    )[:, None]
                )
        return torch.cat(scores, dim=-1)

    def pairwise_scores(
        self, hypotheses: list[str], references: list[str], source: Optional[str] = None
    ) -> Tensor:
        """Calculate the pairwise scores.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.

        Returns:
            Tensor: Score matrix of shape `(H, R)`, where `H` is the number
              of hypotheses and `R` is the number of references.
        """
        with timer.measure("encode/hypotheses"):
            hypotheses_ir = self.encode(hypotheses)
        if hypotheses == references:
            references_ir = hypotheses_ir
        else:
            with timer.measure("encode/references"):
                references_ir = self.encode(references)
        if source is None:
            source_ir = None
        else:
            with timer.measure("encode/source"):
                source_ir = self.encode([source])
        return self.pairwise_scores_from_ir(hypotheses_ir, references_ir, source_ir)


class MetricReferenceless(MetricBase, metaclass=abc.ABCMeta):
    """Base class for reference-less metrics like quality estimation."""

    @abc.abstractmethod
    def score(self, hypothesis: str, source: str) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): A hypothesis.
            source (str): A source.

        Returns:
            float: The score of the given hypothesis.
        """

    def scores(self, hypotheses: list[str], source: str) -> Tensor:
        """Calculate the scores of hypotheses.

        Args:
            hypotheses (list[str]): Hypotheses.
            source (str): A source.

        Returns:
            Tensor: The scores of hypotheses.
        """
        return Tensor([self.score(hyp, source) for hyp in hypotheses])



#!/usr/bin/env python3

import logging
import os
import sys

from mbrs.decoders.base import DecoderBase

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=os.environ.get("LOGLEVEL", "INFO").upper(),
    stream=sys.stdout,
)
logger = logging.getLogger(__name__)

import enum
import json
from argparse import FileType, Namespace
from dataclasses import dataclass

import simple_parsing
from simple_parsing import ArgumentParser, choice, field, flag
from simple_parsing.wrappers import dataclass_wrapper
from tabulate import tabulate, tabulate_formats
from tqdm import tqdm

from mbrs import registry, timer
from mbrs.decoders import DecoderReferenceBased, DecoderReferenceless, get_decoder
from mbrs.metrics import Metric, get_metric

simple_parsing.parsing.logger.setLevel(logging.ERROR)
dataclass_wrapper.logger.setLevel(logging.ERROR)


class Format(enum.Enum):
    plain = "plain"
    json = "json"


@dataclass
class CommonArguments:
    """Common arguments."""

    # Hypotheses file.
    hypotheses: str = field(positional=True)
    # Number of candidates.
    num_candidates: int = field(alias=["-n"])
    # Source file.
    source: str | None = field(default=None, alias=["-s"])
    # References file.
    references: str | None = field(default=None, alias=["-r"])
    # Output file.
    output: FileType("w", encoding="utf-8") = field(default="-", alias=["-o"])
    # Output format.
    format: Format = choice(Format, default=Format.plain)
    # Number of references for each sentence.
    num_references: int | None = field(default=None)
    # Type of the decoder.
    decoder: str = choice(*registry.get_registry("decoder").keys(), default="mbr")
    # Type of the metric.
    metric: str = choice(*registry.get_registry("metric").keys(), default="bleu")
    # Return the n-best hypotheses.
    nbest: int = field(default=1)
    # No verbose information and report.
    quiet: bool = flag(default=False)
    # Report file.
    report: FileType("w") = field(default="-")
    # Report runtime statistics with the given format.
    report_format: str = choice(*tabulate_formats, default="rounded_outline")
    # Number of digits for values of float point.
    width: int = field(default=1, alias=["-w"])


def parse_args() -> Namespace:
    meta_parser = ArgumentParser(add_help=False)
    meta_parser.add_arguments(CommonArguments, "common")
    known_args, _ = meta_parser.parse_known_args()
    metric_type = get_metric(known_args.common.metric)
    decoder_type = get_decoder(known_args.common.decoder)

    parser = ArgumentParser(add_help=True)
    parser.add_arguments(CommonArguments, "common")
    parser.add_arguments(metric_type.Config, "metric", prefix="metric.")
    parser.add_arguments(decoder_type.Config, "decoder", prefix="decoder.")
    return parser.parse_args()


def main(args: Namespace) -> None:
    if not args.common.quiet:
        logger.info(args)

    sources = None
    if args.common.source is not None:
        with open(args.common.source, mode="r") as f:
            sources = f.readlines()

    with open(args.common.hypotheses, mode="r") as f:
        hypotheses = f.readlines()

    references = None
    if args.common.references is not None:
        with open(args.common.references, mode="r") as f:
            references = f.readlines()
    else:
        references = hypotheses

    metric_type = get_metric(args.common.metric)
    metric: Metric = metric_type(args.metric)

    decoder_type = get_decoder(args.common.decoder)
    decoder: DecoderReferenceBased | DecoderReferenceless = decoder_type(
        args.decoder, metric
    )

    num_cands = args.common.num_candidates
    num_refs = args.common.num_references or num_cands
    num_sents = len(hypotheses) // num_cands
    assert num_sents * num_cands == len(hypotheses)

    def output_results(res: DecoderBase.Output):
        if args.common.format == Format.plain:
            for sent in res.sentence:
                print(sent, file=args.common.output)
        elif args.common.format == Format.json:
            for i, (sent, idx, score) in enumerate(
                zip(res.sentence, res.idx, res.score)
            ):
                print(
                    json.dumps(
                        {
                            "rank": i,
                            "sentence": sent,
                            "selected_idx": idx,
                            "expected_score": score,
                        },
                        ensure_ascii=False,
                    ),
                    file=args.common.output,
                )

    if isinstance(decoder, DecoderReferenceless):
        for i in tqdm(range(num_sents)):
            src = sources[i].strip()
            hyps = [h.strip() for h in hypotheses[i * num_cands : (i + 1) * num_cands]]
            with timer.measure("total"):
                output = decoder.decode(hyps, src, args.common.nbest)
            output_results(output)
    else:
        for i in tqdm(range(num_sents)):
            src = sources[i].strip() if sources is not None else None
            hyps = [h.strip() for h in hypotheses[i * num_cands : (i + 1) * num_cands]]
            refs = [r.strip() for r in references[i * num_refs : (i + 1) * num_refs]]
            with timer.measure("total"):
                output = decoder.decode(hyps, refs, src, args.common.nbest)
            output_results(output)

    if not args.common.quiet:
        statistics = timer.aggregate().result(num_sents)
        table = tabulate(
            statistics,
            headers="keys",
            tablefmt=args.common.report_format,
            floatfmt=f".{args.common.width}f",
        )
        print(table, file=args.common.report)


def cli_main():
    args = parse_args()
    main(args)


if __name__ == "__main__":
    cli_main()



import torch

from .cometqe import MetricCOMETQE

SOURCE = "これはテストです"
HYPOTHESES = [
    "this is a test",
    "another test",
    "this is a fest",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
SCORES = torch.Tensor([0.86415, 0.83704, 0.65335, 0.29771])


class TestMetricCOMETQE:
    def test_score(self, metric_cometqe: MetricCOMETQE):
        for i, hyp in enumerate(HYPOTHESES):
            assert torch.isclose(
                SCORES[i],
                torch.tensor(metric_cometqe.score(hyp, SOURCE)),
                atol=0.0005 / 100,
            )

    def test_scores(self, metric_cometqe: MetricCOMETQE):
        scores = metric_cometqe.scores(HYPOTHESES, SOURCE)
        torch.testing.assert_close(
            scores,
            SCORES.to(metric_cometqe.device),
            atol=0.0005 / 100,
            rtol=1e-6,
        )



import time

import pytest

from .timer import ProfileTree, Stopwatch, StopwatchDict


class TestStopwatch:
    def test__init__(self) -> None:
        timer = Stopwatch()
        assert timer._acc_time == 0.0

    def test_reset(self) -> None:
        timer = Stopwatch()
        assert timer._acc_time == 0.0
        timer._acc_time = 1.0
        assert timer._acc_time != 0.0
        timer.reset()
        assert timer._acc_time == 0.0

    def test___call__(self, monkeypatch: pytest.MonkeyPatch):
        t = 0.0

        def perf_counter_mock() -> float:
            return t

        monkeypatch.setattr(time, "perf_counter", perf_counter_mock)

        timer = Stopwatch()
        for _ in range(10):
            with timer():
                t += 1.0
        assert timer.elpased_time == 10.0

    def test_nest_duplicate_measurement(self, monkeypatch: pytest.MonkeyPatch):
        t = 0.0

        def perf_counter_mock() -> float:
            return t

        monkeypatch.setattr(time, "perf_counter", perf_counter_mock)

        timer = Stopwatch()
        with timer():
            t += 10.0
            for _ in range(10):
                with timer():
                    t += 1.0
        assert timer.elpased_time == 10.0

    def test_ncalls(self, monkeypatch: pytest.MonkeyPatch):
        t = 0.0

        def perf_counter_mock() -> float:
            return t

        monkeypatch.setattr(time, "perf_counter", perf_counter_mock)

        timer = Stopwatch()
        for _ in range(20):
            with timer() as ts:
                t += 2.0
                ts.set_delta_ncalls(2)
        assert timer.elpased_time == 40.0
        assert timer.ncalls == 40


class TestStopwatchDict:
    def test__init__(self) -> None:
        timers = StopwatchDict()
        assert len(timers) == 0

    def test_reset(self) -> None:
        timers = StopwatchDict()
        names = ["A", "B"]
        assert all([timers[name]._acc_time == 0.0 for name in names])
        for t in timers.values():
            t._acc_time = 1.0
        assert all([timers[name]._acc_time != 0.0 for name in names])
        timers.reset()
        assert all([timers[name]._acc_time == 0.0 for name in names])

    def test___call__(self, monkeypatch: pytest.MonkeyPatch):
        t = 0.0

        def perf_counter_mock() -> float:
            return t

        monkeypatch.setattr(time, "perf_counter", perf_counter_mock)

        timers = StopwatchDict()
        for _ in range(10):
            with timers("A"):
                t += 1.0
        for i in range(3):
            with timers("B") as ts:
                t += 3.0
                ts.set_delta_ncalls(i + 1)
        assert timers.elapsed_time == {"A": 10.0, "B": 9.0}
        assert timers.ncalls == {"A": 10, "B": 6}

    def test_nest(self, monkeypatch: pytest.MonkeyPatch):
        t = 0.0

        def perf_counter_mock() -> float:
            return t

        monkeypatch.setattr(time, "perf_counter", perf_counter_mock)

        timers = StopwatchDict()
        for _ in range(10):
            with timers("A"):
                for _ in range(3):
                    with timers("B") as ts:
                        t += 3.0
                        ts.set_delta_ncalls(3)
        assert timers.elapsed_time == {"A": 90.0, "B": 90.0}
        assert timers.ncalls == {"A": 10, "B": 90}


class TestProfileTree:
    def test_is_leaf(self):
        root = ProfileTree()
        assert root.is_leaf

        root.children["A"] = ProfileTree()
        root.children["B"] = ProfileTree()
        assert not root.is_leaf
        assert root.children["A"].is_leaf
        assert root.children["B"].is_leaf

    def test_aggregate(self):
        root = ProfileTree()
        for i, name in enumerate(["A", "B", "C"], start=1):
            root.children[name] = ProfileTree(float(i) ** 2, i)
        root.children["D"] = ProfileTree()
        root.children["D"].children["a"] = ProfileTree(7.0, 7)
        root.children["D"].children["b"] = ProfileTree(13.0, 13)
        root.aggregate()
        assert root.elapsed_time == 34.0
        assert root.ncalls == 26

        root = ProfileTree()
        root.children["A"] = ProfileTree(5.0, 5)
        root.children["A"].children["a"] = ProfileTree(7.0, 7)
        root.children["A"].children["b"] = ProfileTree(13.0, 13)
        root.aggregate()
        assert root.elapsed_time == 5.0
        assert root.ncalls == 5
        assert root.children["A"].elapsed_time == 5.0
        assert root.children["A"].ncalls == 5

        root = ProfileTree()
        root.children["A"] = ProfileTree()
        root.children["A"].children["a"] = ProfileTree(ncalls=13)
        with pytest.raises(RuntimeError):
            root.aggregate()
        root = ProfileTree()
        root.children["A"] = ProfileTree()
        root.children["A"].children["a"] = ProfileTree(13.0)
        with pytest.raises(RuntimeError):
            root.aggregate()

    def test_build(self, monkeypatch: pytest.MonkeyPatch):
        t = 0.0

        def perf_counter_mock() -> float:
            return t

        monkeypatch.setattr(time, "perf_counter", perf_counter_mock)

        timers = StopwatchDict()
        for _ in range(10):
            with timers("A"):
                t += 1.0
        for i in range(3):
            with timers("B") as ts:
                t += 3.0
                ts.set_delta_ncalls(i + 1)
        for i in range(5):
            with timers("C/a") as ts:
                t += 3.0
                ts.set_delta_ncalls(3)
            with timers("C/b"):
                t += 1.0

        root = ProfileTree.build(timers)

        assert root.children["A"].elapsed_time == 10.0
        assert root.children["A"].ncalls == 10
        assert root.children["B"].elapsed_time == 9.0
        assert root.children["B"].ncalls == 6
        C = root.children["C"]
        assert C.children["a"].elapsed_time == 15.0
        assert C.children["a"].ncalls == 15
        assert C.children["b"].elapsed_time == 5.0
        assert C.children["b"].ncalls == 5
        assert C.elapsed_time == 20.0
        assert C.ncalls == 20
        assert root.elapsed_time == 39.0
        assert root.ncalls == 36

    def test_result(self, monkeypatch: pytest.MonkeyPatch):
        t = 0.0

        def perf_counter_mock() -> float:
            return t

        monkeypatch.setattr(time, "perf_counter", perf_counter_mock)

        timers = StopwatchDict()
        for _ in range(10):
            with timers("A"):
                t += 1.0
        for i in range(3):
            with timers("B") as ts:
                t += 3.0
                ts.set_delta_ncalls(i + 1)
        for i in range(5):
            with timers("C/a") as ts:
                t += 3.0
                ts.set_delta_ncalls(3)
            with timers("C/b"):
                t += 1.0
            with timers("C/c/1"):
                t += 1.0
            with timers("C/c/2"):
                t += 2.0

        root = ProfileTree.build(timers)
        res = root.result()
        expected = [
            # {"name": "/", "elapsed_time": 54.0, "ncalls": 46, "average_time": 1173.9},
            {"name": "A", "acctime": 10.0, "acccalls": 10, "ms/call": 1000.0},
            {"name": "B", "acctime": 9.0, "acccalls": 6, "ms/call": 1500.0},
            {"name": "C", "acctime": 35.0, "acccalls": 30, "ms/call": 35.0 * 1000 / 30},
            {
                "name": "C/a",
                "acctime": 15.0,
                "acccalls": 15,
                "ms/call": 1000.0,
            },
            {"name": "C/b", "acctime": 5.0, "acccalls": 5, "ms/call": 1000.0},
            {
                "name": "C/c",
                "acctime": 15.0,
                "acccalls": 10,
                "ms/call": 1500.0,
            },
            {
                "name": "C/c/1",
                "acctime": 5.0,
                "acccalls": 5,
                "ms/call": 1000.0,
            },
            {
                "name": "C/c/2",
                "acctime": 10.0,
                "acccalls": 5,
                "ms/call": 2000.0,
            },
        ]
        for r, e in zip(res, expected):
            assert r == e



import pytest
import torch

from .bleu import MetricBLEU

HYPOTHESES = [
    "this is a test",
    "another test",
    "this is a fest",
    "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;",
]
REFERENCES = [
    "ref",
    "this is a test",
    "producţia de zahăr brut se exprimă în zahăr alb;",
]
SCORES_EFFECTIVE_ORDER = {
    True: torch.Tensor(
        [
            [0.0, 100.0, 0.0],
            [0.0, 18.394, 0.0],
            [0.0, 59.460, 0.0],
            [0.0, 0.0, 8.493],
        ]
    ),
    False: torch.Tensor(
        [
            [0.0, 100.0, 0.0],
            [0.0, 0.0, 0.0],
            [0.0, 59.460, 0.0],
            [0.0, 0.0, 8.493],
        ]
    ),
}


class TestMetricBLEU:
    @pytest.mark.parametrize("effective_order", [True, False])
    def test_score(self, effective_order: bool):
        metric = MetricBLEU(MetricBLEU.Config(effective_order=effective_order))
        for i, hyp in enumerate(HYPOTHESES):
            for j, ref in enumerate(REFERENCES):
                assert torch.isclose(
                    SCORES_EFFECTIVE_ORDER[effective_order][i, j],
                    torch.tensor(metric.score(hyp, ref)),
                    atol=0.0005,
                )

    @pytest.mark.parametrize("effective_order", [True, False])
    def test_expected_scores(self, effective_order: bool):
        metric = MetricBLEU(MetricBLEU.Config(effective_order=effective_order))
        expected_scores = metric.expected_scores(HYPOTHESES, REFERENCES)
        torch.testing.assert_close(
            expected_scores,
            SCORES_EFFECTIVE_ORDER[effective_order].mean(dim=1),
            atol=0.0005,
            rtol=1e-4,
        )



from typing import List, Tuple, Union

import pytest
import torch

from .kmeans import Kmeans

N = 100
D = 8
C = 4


def is_equal_shape(
    a: Union[torch.Tensor, torch.Size],
    b: Union[torch.Tensor, torch.Size, List[int], Tuple[int]],
) -> bool:
    """Returns whether a and b have the same shape.

    Args:
        a (Union[torch.Tensor, torch.Size]): An input tensor.
        b (Union[torch.Tensor, torch.Size, List[int], Tuple[int]):
          An input tensor compared the shape with a.

    Returns:
        bool: Whether a and b have the same shape.

    Raises:
        NotImplementedError: When an unsupported type object is given.
    """
    if isinstance(a, torch.Tensor):
        a_shape = a.shape
    elif isinstance(a, torch.Size):
        a_shape = a
    else:
        raise NotImplementedError(f"Type of `a` (`{type(a)}`) is not supported.")

    if isinstance(b, torch.Tensor):
        b_shape = b.shape
    elif isinstance(b, torch.Size):
        b_shape = b
    elif isinstance(b, (list, tuple)):
        b_shape = torch.Size(b)
    else:
        raise NotImplementedError(f"Type of `b` (`{type(b)}`) is not supported.")
    return a_shape == b_shape


class TestKmeans:
    @pytest.mark.parametrize("kmeanspp", [False, True])
    def test___init__(self, kmeanspp: bool):
        kmeans = Kmeans(kmeanspp)
        assert kmeans.kmeanspp == kmeanspp

    def test_init_kmeanspp(self):
        x = torch.rand(N, D)
        kmeans = Kmeans(kmeanspp=True)
        rng = torch.Generator(x.device)
        rng = rng.manual_seed(0)
        centroids = kmeans.init_kmeanspp(x, rng, C)
        assert is_equal_shape(centroids, [C, D])

    def test_assign(self):
        x = torch.rand(N, D)
        centroids = torch.rand(C, D)
        kmeans = Kmeans()
        assigns = kmeans.assign(x, centroids)
        expected = ((x[:, None] - centroids[None, :]) ** 2).sum(dim=-1).argmin(dim=1)
        assert torch.equal(assigns, expected)

    @pytest.mark.parametrize("kmeanspp", [False, True])
    def test_train(self, kmeanspp: bool):
        torch.manual_seed(0)
        kmeans = Kmeans(kmeanspp)
        x = torch.rand(N, D)
        centroids, assigns = kmeans.train(x, C)
        assert is_equal_shape(centroids, [C, D])
        assert is_equal_shape(assigns, [N])



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from mbrs import timer
from mbrs.metrics import MetricCacheable
from mbrs.modules.kmeans import Kmeans

from . import register
from .mbr import DecoderMBR


@register("cbmbr")
class DecoderCBMBR(DecoderMBR):
    """Centroid-Based MBR decoder class.

    - Time complexity: O(Nk)
    - Space complexity: O(Nk)
    where k << N.
    """

    def __init__(self, cfg: Config, metric: MetricCacheable) -> None:
        super().__init__(cfg, metric)
        self.kmeans = Kmeans(kmeanspp=self.cfg.kmeanspp)

    cfg: Config

    @dataclass
    class Config(DecoderMBR.Config):
        """Configuration for the decoder.

        - ncentroids (int): Number of centroids.
        - niter (int): Number of k-means iteration
        - count_weight: (bool) Weight the scores with counts.
        - kmeanspp (bool): Initialize the centroids using k-means++.
        - seed (bool): Random seed.
        """

        ncentroids: int = 8
        niter: int = 3
        count_weight: bool = False
        kmeanspp: bool = True
        seed: int = 0

    def decode(
        self,
        hypotheses: list[str],
        references: list[str],
        source: Optional[str] = None,
        nbest: int = 1,
    ) -> DecoderCBMBR.Output:
        """Select the n-best hypotheses based on the strategy.

        Args:
            hypotheses (list[str]): Hypotheses.
            references (list[str]): References.
            source (str, optional): A source.
            nbest (int): Return the n-best hypotheses.

        Returns:
            DecoderCBMBR.Output: The n-best hypotheses.
        """

        assert isinstance(self.metric, MetricCacheable)

        with timer.measure("encode/hypotheses"):
            hypotheses_ir = self.metric.encode(hypotheses)
        if hypotheses == references:
            references_ir = hypotheses_ir
        else:
            with timer.measure("encode/references"):
                references_ir = self.metric.encode(references)
        if source is None:
            source_ir = None
        else:
            with timer.measure("encode/source"):
                source_ir = self.metric.encode([source])
        centroids, assigns = self.kmeans.train(
            references_ir,
            min(self.cfg.ncentroids, len(references)),
            niter=self.cfg.niter,
            seed=self.cfg.seed,
        )
        if self.cfg.count_weight:
            centroid_ids, counts_nonzero = assigns.unique(return_counts=True)
            counts = centroid_ids.new_zeros(len(centroids))
            counts[centroid_ids] = counts_nonzero
            counts = counts.float()
            weights = counts / counts.sum()
            with timer.measure("expectation"):
                pairwise_scores = self.metric.pairwise_scores_from_ir(
                    hypotheses_ir, centroids, source_ir
                )
                expected_scores = (pairwise_scores * weights[None, :]).mean(dim=-1)
        else:
            with timer.measure("expectation"):
                expected_scores = self.metric.pairwise_scores_from_ir(
                    hypotheses_ir, centroids, source_ir
                ).mean(dim=-1)

        topk_scores, topk_indices = self.metric.topk(expected_scores, k=nbest)
        return self.Output(
            idx=topk_indices,
            sentence=[hypotheses[idx] for idx in topk_indices],
            score=topk_scores,
        )



from __future__ import annotations

from dataclasses import dataclass

from sacrebleu.metrics.ter import TER

from . import Metric, register


@register("ter")
class MetricTER(Metric):
    """TER metric class."""

    HIGHER_IS_BETTER: bool = False

    @dataclass
    class Config(Metric.Config):
        """TER metric configuration.

        - normalized (bool): Enable character normalization.
          By default, normalizes a couple of things such as newlines being stripped,
          retrieving XML encoded characters, and fixing tokenization for punctuation.
          When 'asian_support' is enabled, also normalizes specific Asian (CJK)
          character sequences, i.e. split them down to the character level.
        - no_punct (bool): Remove punctuation. Can be used in conjunction with
          'asian_support' to also remove typical punctuation markers in Asian languages
          (CJK).
        - asian_support (bool): Enable special treatment of Asian characters.
          This option only has an effect when 'normalized' and/or 'no_punct' is enabled.
          If 'normalized' is also enabled, then Asian (CJK) characters are split down to
          the character level. If 'no_punct' is enabled alongside 'asian_support',
          specific unicode ranges for CJK and full-width punctuations are also removed.
        - case_sensitive (bool): If `True`, does not lowercase sentences.
        """

        normalized: bool = False
        no_punct: bool = False
        asian_support: bool = False
        case_sensitive: bool = False

    def __init__(self, cfg: MetricTER.Config):
        self.scorer = TER(
            normalized=cfg.normalized,
            no_punct=cfg.no_punct,
            asian_support=cfg.asian_support,
            case_sensitive=cfg.case_sensitive,
        )

    def score(self, hypothesis: str, reference: str, *_) -> float:
        """Calculate the score of the given hypothesis.

        Args:
            hypothesis (str): Hypothesis.
            reference (str): Reference.

        Returns:
            float: The score of the given hypothesis.
        """
        return self.scorer.sentence_score(hypothesis, [reference]).score
